//***************************************************************************
// Group Linslot / Linux - Slotrace Manager
// File arduino.cc
// Date 16.11.08 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//***************************************************************************

//***************************************************************************
// Include
//***************************************************************************

#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>

#include <arduino.hpp>

//***************************************************************************
// 
//***************************************************************************

const char* cmdGetTime = "GETTIME;";

//***************************************************************************
// Object
//***************************************************************************

Arduino::Arduino()
   : IoInterface()
{
   fdDevice = 0;
   readTimeout = 60;
   writeTimeout = 60;

   tvNow(&tvBoarStartTime);
   tvNow(&tvLastIoTime);

   bzero(&oldtio, sizeof(oldtio));

   outValue = 0xFFFFFFFF;
}

Arduino::~Arduino()
{
   close();
}

//***************************************************************************
// Open/Close Device
//***************************************************************************

int Arduino::open(const char* dev)
{
   struct termios newtio;

   if (!dev)
      dev = deviceName;
   else
      strcpy(deviceName, dev);

   if (!deviceName || !*deviceName)
      return fail;

   QMutexLocker locker(&deviceMutex);

   if ((fdDevice = ::open(deviceName, O_RDWR | O_NOCTTY)) < 0)
   {
      tell(eloAlways, "Error: Opening '%s' failed", deviceName);
      return fail;
   }

   // configure serial line

   tcgetattr(fdDevice, &oldtio);
   bzero(&newtio, sizeof(newtio));
   newtio.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
   newtio.c_iflag = IGNPAR | PARMRK;
   newtio.c_oflag = 0;
   newtio.c_lflag = 0;
   tcflush(fdDevice, TCIFLUSH);
   tcsetattr(fdDevice, TCSANOW, &newtio);

   // first ask about the boardtime

   write(fdDevice, cmdGetTime, strlen(cmdGetTime));

   return success;
}

int Arduino::close()
{
   // restore the old settings and close

   if (isOpen())
   {
      tell(eloDebug, "Debug: Closing connection");

      QMutexLocker locker(&deviceMutex);

      tcsetattr(fdDevice, TCSANOW, &oldtio);
      ::close(fdDevice);
      fdDevice = 0;
   }

   return success;
}

//***************************************************************************
// Re Open
//***************************************************************************

int Arduino::reopen(const char* dev)
{
   close();
   
   return open(dev);
}

int Arduino::flush()
{
   return success;
}

//***************************************************************************
// Write IO Bit
//***************************************************************************

int Arduino::writeBit(int bit, int state)
{
   char* cmd;

   if (!isOpen() || bit == na)
      return fail;

   setBitTo(outValue, bit, state);
   asprintf(&cmd, "DO%d=%d;", bit, state);

   tell(eloDebug, "Debug: Write bit (%d) value (%d) [%s]", 
        bit, state, cmd);

   QMutexLocker locker(&deviceMutex);
   ::write(fdDevice, cmd, strlen(cmd));

   free(cmd);

   return success;
}

//***************************************************************************
// Read Out Bit
//
//  - Return the actual Value of the given 'out' Bit,
//    using the cached value of last write!
//***************************************************************************

int Arduino::readOutBit(int bit)
{
   char buf[100+TB];
   
   tell(eloDebug2, "Debug: out value is '%s'", toBinStr(outValue, buf));

   return isBit(outValue, bit);
}

//***************************************************************************
// ...
//***************************************************************************

int Arduino::writeByte(unsigned char /*byte*/, int /*byteNum*/)
{
   // not needed ... ?

   return success;
}

//***************************************************************************
// Read Value
//***************************************************************************

int Arduino::readValue(unsigned int* value)
{
   char b;
   char cmd[sizeCmdMax+TB];
   IoValue iov;

   if (!isOpen())
      return fail;

   deviceMutex.lock();

   if (::read(fdDevice, &b, 1) <= 0)
   {
      // no input data pending

      deviceMutex.unlock();
      return fail;
   }

   // tell(eloDebug2, "Debug: Got lookahead byte '0x%X'", b);
   deviceMutex.unlock();
   readCmd(cmd, b);

   if (dispatch(cmd, iov) != cmdDigitalInput)
      return ignore;

   if (value)
      *value = iov.dword;

   return success;
}

//***************************************************************************
// 
//***************************************************************************

int Arduino::lookByte(int /*byteNum*/, unsigned  char* /*byte*/)
{
   // not needed ... ?

   return success;
}

//***************************************************************************
// Settings
//***************************************************************************

int Arduino::setTimeout(unsigned int timeout)
{
   readTimeout = timeout / 1000;
   return success;
}

int Arduino::setWriteTimeout(unsigned int timeout)
{
   writeTimeout = timeout / 1000;
   return success;
}

//***************************************************************************
// Read Command
//***************************************************************************

const char* Arduino::readCmd(char* command, char lookahead)
{
   int count = 0;
   char b;
   int start = time(0);
   int cnt;

   *command = 0;

   if (!isOpen())
      return command;

   if (lookahead && lookahead != '\n')
   {
      command[0] = lookahead;
      command[1] = 0;
      count++;
   }

   // QMutexLocker locker(&deviceMutex);

   while (time(0) < start + readTimeout)
   { 
      deviceMutex.lock();
      cnt = ::read(fdDevice, &b, 1);
      deviceMutex.unlock();

      if (cnt <= 0)
      {
         usleep(100);
         continue;
      }

      // skip cartridge return 

      if (b == '\r')
         continue;

      // stop at linefeed

      if (b == '\n')
         break;

      command[count] = b;
      count++;      
   }

   command[count] = 0;

   tell(eloDebug2, "Debug: <- (%d)[%s]", count, command);

   return command;
}

//***************************************************************************
// Dispatch
//***************************************************************************

int Arduino::dispatch(char* buf, IoValue& value)
{
   // Examples:
   //     "BOARDTIME=750205;"
   //     "DI:754041=2999;"

   char* p = buf;
   char* cmd = buf;
   char* val = 0;
   char* t = 0;

   if ((p = strchr(p, ':')))
   {
      *p = 0;
      p++;
      t = p;
   }
   else
      p = buf;

   if (!(p = strchr(p, '=')))
      return fail;

   *p = 0;
   p++;
   val = p;

   if (!(p = strchr(p, ';')))
      return fail;

   *p = 0;

   // the value can transmitted in hex or dec

   if (strncasecmp(val, "0x", 2)  == 0)
      value.dword = strtoll(val+2, 0, 16);
   else
      value.dword = atoll(val);

   // save the time

   if (t) 
      tvLastIoTime = addMs2Tv(tvBoarStartTime, atoll(t));

   // some debug stuff

   char tmp[100+TB];

   tell(eloDebug2, "Debug: Got command '%s' - '%s' at (%sms)",
        notNull(cmd), toBinStr(atoi(val), tmp), notNull(t));

   // at least dispatch the command

   if (strcasecmp(cmd, "BOARDTIME") == 0)
   {
      tvNow(&tvBoarStartTime);
      tvBoarStartTime = addMs2Tv(tvBoarStartTime, atoll(val));
      return cmdBoardTime;
   }

   if (strcasecmp(cmd, "DI") == 0)
      return cmdDigitalInput;

   if (strcasecmp(cmd, "AI") == 0)
      return cmdAnalogInput;

   return fail;
}
